-- ==============================================================
-- FILE : PLB2DDR_RD.VHD
-- ==============================================================
--
-- This file is part of the PLB Power PC to DDR interface for the
-- HERON-FPGA9.
--
-- This file is derived from Xilinx provided example code.
--
-- ==============================================================
--
--  Date        : 16/09/2005
--  Author      : R. Williams - HUNT ENGINEERING
--  Description : PLB to DDR interface read logic
--
-- ==============================================================
--
--  Ver     Modified By      Date      Changes
--  ---     -----------      ----      -------
--  1.0     R. Williams    16-09-05    First Written
--
-- ==============================================================


library IEEE;
  use IEEE.std_logic_1164.all;
  use IEEE.std_logic_arith.all;
  use IEEE.std_logic_unsigned.all;

-- pragma translate_off
library UNISIM;
  use UNISIM.vcomponents.all;
-- pragma translate_on


entity PLB2DDR_RD is
  port (
    Bus2IP_UserClk   : in  std_logic;
    Bus2IP_UserRst   : in  std_logic;
    Bus2IP_Addr      : in  std_logic_vector(0 to 31);
    Bus2IP_BE        : in  std_logic_vector(0 to 7);
    Bus2IP_Burst     : in  std_logic;
    Bus2IP_CE        : in  std_logic_vector(0 to 0);
    Bus2IP_CS        : in  std_logic_vector(0 to 0);
    Bus2IP_RNW       : in  std_logic;
    Bus2IP_RdCE      : in  std_logic_vector(0 to 0);
    Bus2IP_RdReq     : in  std_logic;
    IP2Bus_Data      : out std_logic_vector(0 to 63);
    IP2Bus_RdAddrAck : out std_logic;
    IP2Bus_RdAck     : out std_logic;
    RD_DATA          : in  std_logic_vector(63 downto 0);
    RD_ADDR          : out std_logic_vector(24 downto 0);
    RD_ADDR_WEN      : out std_logic;
    RD_ADDR_FF       : in  std_logic;
    RD_ADDR_AF       : in  std_logic;
    RD_RISE_REN      : out std_logic;
    RD_RISE_EF       : in  std_logic;
    RD_RISE_AE       : in  std_logic;
    RD_FALL_REN      : out std_logic;
    RD_FALL_EF       : in  std_logic;
    RD_FALL_AE       : in  std_logic
  );
end PLB2DDR_RD;


architecture RTL of PLB2DDR_RD is

  signal Read_Enable      : std_logic;
  signal Read_Allow       : std_logic;
  signal Read_Addr_Toggle : std_logic;
  signal Read_DelayedReq  : std_logic;
  signal Read_Done        : std_logic;
  signal Read_RiseRen     : std_logic;
  signal Read_FallRen     : std_logic;

begin

  IP2Bus_RdAddrAck <= Read_Enable;
  IP2Bus_RdAck     <= Read_Done AND Bus2IP_CE(0);

  RD_ADDR <= Bus2IP_Addr(5 to 28) & '0';

  RD_ADDR_WEN <= Read_Enable AND (NOT Read_Addr_Toggle);

  Read_Enable <= Read_Allow AND (Bus2IP_RdReq OR Read_DelayedReq);

  Read_Allow <= (NOT RD_ADDR_AF);

  process(Bus2IP_UserRst, Bus2IP_UserClk)
  begin
    if Bus2IP_UserRst='1' then
      Read_DelayedReq <= '0';
    elsif rising_edge(Bus2IP_UserClk) then
      Read_DelayedReq <= '0';
      if (Bus2IP_RdReq='1'    and Read_Allow='0') or
         (Read_DelayedReq='1' and Read_Allow='0') then
        Read_DelayedReq <= '1';
      end if;
    end if;
  end process;

  process(Bus2IP_UserRst, Bus2IP_UserClk)
  begin
    if Bus2IP_UserRst='1' then
      Read_Addr_Toggle <= '0';
    elsif rising_edge(Bus2IP_UserClk) then
      Read_Addr_Toggle <= '0';
      if (Read_Addr_Toggle='0' and Read_Enable='1' and Bus2IP_CE(0)='1') or
         (Read_Addr_Toggle='1' and Read_Enable='0' and Bus2IP_CE(0)='1') then
        Read_Addr_Toggle <= '1';
      end if;
    end if;
  end process;

  RD_RISE_REN <= Read_RiseRen;
  RD_FALL_REN <= Read_FallRen;

  Read_RiseRen <= (NOT RD_RISE_EF) AND (NOT RD_FALL_EF);
  Read_FallRen <= (NOT RD_RISE_EF) AND (NOT RD_FALL_EF);

  process(Bus2IP_UserRst, Bus2IP_UserClk)
  begin
    if Bus2IP_UserRst='1' then
      Read_Done <= '0';
    elsif rising_edge(Bus2IP_UserClk) then
      Read_Done <= '0';
      if Read_RiseRen='1' then
        Read_Done <= '1';
      end if;
    end if;
  end process;

  IP2Bus_Data(0 to 31)  <= RD_DATA(31 downto 0);
  IP2Bus_Data(32 to 63) <= RD_DATA(63 downto 32);

end RTL;
