--SINGLE_FILE_TAG
-------------------------------------------------------------------------------
-- $Id: pwm.vhd,v 1.2 2003/06/27 00:56:18 anitas Exp $
-------------------------------------------------------------------------------
-- PWM - entity/architecture pair
-------------------------------------------------------------------------------
--
--  ***************************************************************************
--  **  Copyright(C) 2003 by Xilinx, Inc. All rights reserved.               **
--  **                                                                       **
--  **  This text contains proprietary, confidential                         **
--  **  information of Xilinx, Inc. , is distributed by                      **
--  **  under license from Xilinx, Inc., and may be used,                    **
--  **  copied and/or disclosed only pursuant to the terms                   **
--  **  of a valid license agreement with Xilinx, Inc.                       **
--  **                                                                       **
--  **  Unmodified source code is guaranteed to place and route,             **
--  **  function and run at speed according to the datasheet                 **
--  **  specification. Source code is provided "as-is", with no              **
--  **  obligation on the part of Xilinx to provide support.                 **
--  **                                                                       **
--  **  Xilinx Hotline support of source code IP shall only include          **
--  **  standard level Xilinx Hotline support, and will only address         **
--  **  issues and questions related to the standard released Netlist        **
--  **  version of the core (and thus indirectly, the original core source). **
--  **                                                                       **
--  **  The Xilinx Support Hotline does not have access to source            **
--  **  code and therefore cannot answer specific questions related          **
--  **  to source HDL. The Xilinx Support Hotline will only be able          **
--  **  to confirm the problem in the Netlist version of the core.           **
--  **                                                                       **
--  **  This copyright and support notice must be retained as part           **
--  **  of this text at all times.                                           **
--  ***************************************************************************
--
-------------------------------------------------------------------------------
-- Filename:        pwm.vhd
-- Version:         v1.00a
-- Description:     Pulse width modulator
--
-------------------------------------------------------------------------------
-- Structure: 
--
--              pwm.vhd
--
-------------------------------------------------------------------------------
-- Author:      BLT
-- History:
--  BLT             10-22-2002      -- First version
-- ^^^^^^
--      First version of pwm.
-- ~~~~~~
-------------------------------------------------------------------------------
-- Naming Conventions:
--      active low signals:                     "*_n"
--      clock signals:                          "clk", "clk_div#", "clk_#x" 
--      reset signals:                          "rst", "rst_n" 
--      generics:                               "C_*" 
--      user defined types:                     "*_TYPE" 
--      state machine next state:               "*_ns" 
--      state machine current state:            "*_cs" 
--      combinatorial signals:                  "*_cmb" 
--      pipelined or register delay signals:    "*_d#" 
--      counter signals:                        "*cnt*"
--      clock enable signals:                   "*_ce" 
--      internal version of output port         "*_i"
--      device pins:                            "*_pin" 
--      ports:                                  - Names begin with Uppercase 
--      processes:                              "*_PROCESS" 
--      component instantiations:               "<ENTITY_>I_<#|FUNC>
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_signed.all;
use IEEE.std_logic_misc.all;

library Unisim;
use Unisim.Vcomponents.all;


-------------------------------------------------------------------------------
-- Port Declarations
-------------------------------------------------------------------------------
-- Definition of Ports:
--      Clk                         -- clock                                             
--      Control                     -- PWM setting
--      PWM_out                     -- output
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-- Entity section
-------------------------------------------------------------------------------
entity PWM is
  port (
    Clk         : in  std_logic;
    Control     : in  std_logic_vector(0 to 7);
    PWM_Out     : out std_logic
    );
end entity PWM;

-------------------------------------------------------------------------------
-- Architecture section
-------------------------------------------------------------------------------

architecture IMP of PWM is

-------------------------------------------------------------------------------
-- Signal declarations
-------------------------------------------------------------------------------
signal srl1_q    : std_logic;
signal srl1_q15  : std_logic;
signal srl2_q    : std_logic;
signal srl2_q15  : std_logic;
signal fdc_reset : std_logic;

-------------------------------------------------------------------------------
-- Attributes
-------------------------------------------------------------------------------
attribute INIT : string;
attribute INIT of SRL1_I : label is "0001";
attribute INIT of SRL2_I : label is "FFFE";

-------------------------------------------------------------------------------
-- Begin architecture
-------------------------------------------------------------------------------

begin -- architecture IMP

SRL1_I: SRLC16E
-- synthesis translate_off
  generic map (
    INIT => X"0001"
  )
-- synthesis translate_on
  port map (
    Q    =>  srl1_q,
    Q15  =>  srl1_q15,
    CLK  =>  Clk,
    CE   =>  '1',
    A0   =>  Control(7),
    A1   =>  Control(6),
    A2   =>  Control(5),
    A3   =>  Control(4),
    D    =>  srl1_q15
  );
  
SRL2_I: SRLC16E
-- synthesis translate_off
  generic map (
    INIT => X"FFFE"
  )
-- synthesis translate_on
  port map (
    Q    =>  srl2_q,
    Q15  =>  srl2_q15,
    CLK  =>  Clk,
    CE   =>  srl1_q15,
    A0   =>  Control(3),
    A1   =>  Control(2),
    A2   =>  Control(1),
    A3   =>  Control(0),
    D    =>  srl2_q15
  );
  
fdc_reset <= srl1_q and not srl2_q;

FDC_I: FDC
  port map (
    Q    =>  PWM_Out,
    C    =>  srl2_q15,
    CLR  =>  fdc_reset,
    D    =>  '1'
  );
   
end architecture IMP;
